#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os
import errno
from cliboa.util.lisboa_log import LisboaLog
from datetime import datetime
from time import sleep
from ftplib import FTP, FTP_TLS


class FtpUtil(object):
    """
    FTP common operation
    """

    TIMEOUT_SEC = 30
    RETRY_SEC = 10

    def __init__(
        self,
        host,
        user,
        password,
        timeout=TIMEOUT_SEC,
        retryTimes=3,
        port=21,
        tls=False,
    ):
        """
        Must set whether password or key

        Args:
            host (str): hostname
            user (str): username
            password (str): password
            timeout=30 (int): timeout seconds
            retryTimes=3 (int): retry count
            port=21 (int): port number
            tls=False (bool): use secure connection
        """

        self.__host = host
        self.__user = user
        self.__password = password
        self.__timeout = timeout
        self.__retryTimes = retryTimes
        self.__port = port
        self.__tls = tls
        self.__logger = LisboaLog.get_logger(__name__)

    def list_files(self, dir, dest, pattern):
        """
        Function that is to download files by regular pattern matching

        Args:
            dir (str): ftp server directory to download
            dest (str): local directory where to place downloaded files from ftp
            pattern (object): regular pattern matching

        Returns:
            list: downloaded file names

        Raises:
            IOError: ftplib failure
        """
        return self.__execute(list_file_func, dir=dir, dest=dest, pattern=pattern)

    def clear_files(self, dir, pattern):
        """
        Remove all files which matche to pattern. Not remove directory.

        Args:
            dir (str): remove target dir
            pattern (object): remove target file pattern

        Raises:
            IOError: failed to remove
        """
        self.__execute(clear_file_func, dir=dir, pattern=pattern)

    def remove_specific_file(self, dir, fname):
        """
        Remove specific file (by name) from target directory

        Args:
            dir (str): remove target dir
            fname (str): file name (exact match)

        Raises:
            IOError: failed to remove
        """
        self.__execute(remove_specific_file_func, dir=dir, fname=fname)


    def file_mdtm(self, dir, unixtime=False):
        """
        Returns dictionary of file name and timestamp from target directory

        Args:
            dir (str): directory
            unixtime=False (bool): Whether response time returns as str(yyyyMMddHHmmss) or unixtime
        Returns:
            dict: {file name: update time}


        Raises:
            IOError: ftplib failure
        """
        return self.__execute(file_mdtm_func, dir=dir, unixtime=unixtime)

    def __execute(self, func, **kwargs):
        """
        Connect to ftp server and execute defined function.

        Args:
            func (fucntion): Callback function after connected to the ftp server.
            kwargs (dict): arguments
                           key name 'ftp' is a reserved word. do NOT use it.

        Raises:
            ValueError: Arguments are empty
            IOError: ftplib failure
        """

        if not func:
            raise ValueError("Function must not be empty.")

        for _ in range(self.__retryTimes):
            try:
                return self.__ftp_call(func, **kwargs)
            except Exception as e:
                self.__logger.warning(e)
                self.__logger.warning(kwargs)

            self.__logger.warning(
                "Unexpected error occurred. Retry will start in 10 sec."
            )
            sleep(FtpUtil.RETRY_SEC)

        raise IOError(errno.ENOENT, "FTP failed.")

    def __ftp_call(self, func, **kwargs):
        if self.__tls:
            with FTP_TLS(host=self.__host, timeout=self.__timeout) as ftp:
                ftp.set_debuglevel(1)
                ftp.login(user=self.__user, passwd=self.__password)
                ftp.prot_p()
                return func(ftp=ftp, **kwargs)
        else:
            with FTP() as ftp:
                ftp.set_debuglevel(1)
                ftp.connect(host=self.__host, port=self.__port, timeout=self.__timeout)
                ftp.login(user=self.__user, passwd=self.__password)
                return func(ftp=ftp, **kwargs)


def list_file_func(**kwargs):
    files = []
    for src in kwargs["ftp"].nlst(kwargs["dir"]):
        fname = os.path.basename(src)
        if kwargs["pattern"].match(fname) is None:
            continue

        fPath = os.path.join(kwargs["dir"], src)
        try:
            with open(os.path.join(kwargs["dest"], fname), "wb") as f:
                kwargs["ftp"].retrbinary("RETR " + src, f.write)
            files.append(fname)
        except IOError:
            # ignore error. try to get all files
            pass
    return files


def clear_file_func(**kwargs):
    for src in kwargs["ftp"].nlst(kwargs["dir"]):
        fname = os.path.basename(src)
        if kwargs["pattern"].match(fname) is None:
            continue
        try:
            kwargs["ftp"].delete(src)
        except Exception as e:
            # ignore erros. Directory cannot be deleted by ftp#delete.
            pass


def remove_specific_file_func(**kwargs):
    kwargs["ftp"].delete(os.path.join(kwargs["dir"], kwargs["fname"]))


def file_mdtm_func(**kwargs):
    res = {}
    for src in kwargs["ftp"].nlst(kwargs["dir"]):
        mdtm = kwargs['ftp'].voidcmd('MDTM %s' % src)[4:].strip()
        if kwargs['unixtime'] is True:
            mdtm = int(datetime.strptime(mdtm, '%Y%m%d%H%M%S').timestamp())
        res[os.path.basename(src)] = mdtm
    return res



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import ast
import codecs
import csv

from cliboa.scenario.base import BaseSqlite
from cliboa.scenario.validator import (
    EssentialParameters,
    SqliteTableExistence,
    IOOutput,
)
from cliboa.util.exception import SqliteInvalid, FileNotFound


class SqliteTransaction(BaseSqlite):
    """
    Base class of sqlite transacdtion
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        """
        Execute db connection, commit, close, vacuum

        Args:
            args[0]: db operation function to execute as transaction
        """

        # When args is not specified, just call validation of base class
        if not args:
            super().execute()
            return 1
        func = args[0]

        # TODO
        # Add to handle to select DB transaction seperation level
        self._logger.info("Start DB Transaction")
        self._sqlite_adptr.connect(self._dbname)
        try:
            func()
            self._sqlite_adptr.commit()
        finally:
            super()._close_database()
        self._logger.info("Finish DB Transaction")


class SqliteCreation(SqliteTransaction):
    """
    Insert all the input data specified as 'io: input' in yaml to the specified table.
    """

    def __init__(self):
        super().__init__()
        self._tblname = None
        self._replace_into = True
        self._insert_cnt = 10
        self._primary_key = None
        self._refresh = False

    @property
    def tblname(self):
        return self._tblname

    @tblname.setter
    def tblname(self, tblname):
        self._tblname = tblname

    @property
    def replace_into(self):
        return self._replace_into

    @replace_into.setter
    def replace_into(self, replace_into):
        self._replace_into = replace_into

    @property
    def insert_cnt(self):
        return self._insert_cnt

    @insert_cnt.setter
    def insert_cnt(self, insert_cnt):
        self._insert_cnt = insert_cnt

    @property
    def primary_key(self):
        return self._primary_key

    @primary_key.setter
    def primary_key(self, primary_key):
        self._primary_key = primary_key

    @property
    def refresh(self):
        return self._refresh

    @refresh.setter
    def refresh(self, refresh):
        self._refresh = refresh

    def execute(self, *args):
        super().execute()
        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))

        param_valid = EssentialParameters(self.__class__.__name__, [self._tblname])
        param_valid()

        tbl_valid = SqliteTableExistence(self._dbname, self._tblname)
        tbl_valid()

        output_valid = IOOutput(self._io)
        output_valid()

        # get table column definition
        self._sqlite_adptr.connect(self._dbname)
        column_def = self.__get_column_def()

        if self._refresh is True:
            self.__refresh_table(column_def)

        # database transaction
        def insert():
            self._logger.info("Start to insert")
            insert_rows = []
            with open(self._s.cache_file, "r", encoding="utf-8") as f:
                for i, l_str in enumerate(f, 1):
                    l_dict = ast.literal_eval(l_str)
                    insert_rows.append(l_dict)

                    # Check only once
                    if i == 1:
                        self.__valid_column_def(column_def, l_dict)

                    # execute bulk insert
                    if i % self._insert_cnt == 0:
                        self._sqlite_adptr.execute_many_insert(
                            self._tblname, column_def, insert_rows, self._replace_into
                        )
                        insert_rows.clear()

                if len(insert_rows) > 0:
                    self._sqlite_adptr.execute_many_insert(
                        self._tblname, column_def, insert_rows, self._replace_into
                    )
                    insert_rows.clear()

            self._logger.info("Finish to insert")

        super().execute(insert)
        self._s.remove()

    def __get_column_def(self):
        """
        Get table column defenition
        """
        if self._columns:
            return self._columns

        result = self._sqlite_adptr.fetch("PRAGMA TABLE_INFO(%s)" % self._tblname)
        column_def = []
        for c in result:
            column_def.append(c[1])
        return column_def

    def __valid_column_def(self, column_def, l_dict):
        """
        Check column definitions in scenario.yml and cache file columns
        """
        l_keys_list = list(l_dict.keys())
        if column_def != l_keys_list:
            inconsistent_columns = set(column_def) ^ set(l_keys_list)
            raise SqliteInvalid(
                "The columns %s is not consistent between scenario.yml and input cache."
                % ", ".join(inconsistent_columns)
            )

    def __refresh_table(self, column_def):
        """
        Drop and recreate the table by using the existing table column definition
        """

        # def drop_and_create_tbl():
        self._logger.info("Drop table %s." % self._tblname)
        self._sqlite_adptr.execute("DROP TABLE IF EXISTS %s" % self._tblname)
        self._logger.info("Create new table %s." % self._tblname)

        if self._primary_key is None:
            sql = "CREATE TABLE IF NOT EXISTS %s (%s)"
            self._sqlite_adptr.execute(
                sql % (self._tblname, " TEXT, ".join(column_def) + " TEXT")
            )
        else:
            sql = "CREATE TABLE IF NOT EXISTS %s (%s, PRIMARY KEY(%s))"
            self._sqlite_adptr.execute(
                sql
                % (
                    self._tblname,
                    " TEXT, ".join(column_def) + " TEXT",
                    self._primary_key,
                )
            )



class CsvReadSqliteCreate(SqliteTransaction):

    COMMIT_COUNT = 100

    def __init__(self):
        super().__init__()
        self._src_dir = None
        self._src_pattern = None
        self._tblname = None
        self._primary_key = None
        self._index = []
        self._refresh = True
        self._encoding = 'utf-8'

    @property
    def src_dir(self):
        return self._src_dir

    @src_dir.setter
    def src_dir(self, src_dir):
        self._src_dir = src_dir

    @property
    def src_pattern(self):
        return self._src_pattern

    @src_pattern.setter
    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    @property
    def tblname(self):
        return self._tblname

    @tblname.setter
    def tblname(self, tblname):
        self._tblname = tblname

    @property
    def primary_key(self):
        return self._primary_key

    @primary_key.setter
    def primary_key(self, primary_key):
        self._primary_key = primary_key

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, index):
        self._index = index

    @property
    def refresh(self):
        return self._refresh

    @refresh.setter
    def refresh(self, refresh):
        self._refresh = refresh

    @property
    def encoding(self):
        return self._encoding

    @encoding.setter
    def encoding(self, encoding):
        self._encoding = encoding

    def execute(self, *args):
        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))

        # essential parameters check
        valid = EssentialParameters(self.__class__.__name__, [self._src_dir, self.src_pattern, self._tblname])
        valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        self._logger.info("Files found %s" % files)

        if len(files) > 1:
            raise Exception('Input file must be only one.')

        if len(files) == 0:
            raise FileNotFound('No csv file was found.')

        def func():
            if self._refresh is True:
                # Drop table in advance, If refresh is True
                self._sqlite_adptr.execute("DROP TABLE IF EXISTS %s" % self._tblname)
                self._sqlite_adptr.commit()

            with codecs.open(files[0], mode="r", encoding=self._encoding) as f:
                reader = csv.DictReader(f)
                # Table columns will be the same with csv column names.
                escaped_columns = ['"%s"' % fn for fn in reader.fieldnames]

                self._logger.info("Create table [%s]" % self._tblname)
                if self._primary_key is None:
                    sql = "CREATE TABLE IF NOT EXISTS %s (%s)"
                    self._sqlite_adptr.execute(sql % (self._tblname, " TEXT, ".join(escaped_columns) + " TEXT"))
                else:
                    sql = "CREATE TABLE IF NOT EXISTS %s (%s, PRIMARY KEY(%s))"
                    self._sqlite_adptr.execute(sql % (self._tblname, " TEXT, ".join(escaped_columns) + " TEXT", self._primary_key))
                self._sqlite_adptr.commit()

                # Put all csv records into the table.
                self._logger.info("Insert all csv records into table[%s]" % self._tblname)
                params = []
                for row in reader:
                    params.append(row)
                    if len(params) == self.COMMIT_COUNT:
                        self._sqlite_adptr.execute_many_insert(self._tblname, reader.fieldnames, params, False)
                        self._sqlite_adptr.commit()
                        params.clear()
                if len(params) > 0:
                    self._sqlite_adptr.execute_many_insert(self._tblname, reader.fieldnames, params, False)
                    self._sqlite_adptr.commit()

            if self._index and len(self._index) > 0:
                # Create index (Add the index at the end for better performance when insert data is large)
                self._logger.info("Add index")
                self._sqlite_adptr.add_index(self._tblname, self._index)
                self._sqlite_adptr.commit()

        super().execute(func)



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import fnmatch
import os
import pandas
import random
import re
import string
from datetime import datetime
from google.cloud import storage, bigquery

from cliboa.scenario.validator import EssentialParameters
from cliboa.scenario.gcp import BaseBigQuery, BaseGcs
from cliboa.util.cache import ObjectStore
from cliboa.util.gcp import ServiceAccount


class BigQueryReadCache(BaseBigQuery):
    """
    Get data from BigQuery and cache them as pandas.dataframe format.

    Use {BigQueryFileDownload} if the result query is estimated to be large.
    """

    def __init__(self):
        super().__init__()
        self._key = None
        self._query = None

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, key):
        self._key = key

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query):
        self._query = query

    def _get_query(self):
        if self._query is None:
            return "SELECT * FROM %s.%s" % (self._dataset, self._tblname)
        else:
            return self._query

    def execute(self, *args):
        super().execute()
        valid = EssentialParameters(self.__class__.__name__, [self._key])
        valid()

        df = pandas.read_gbq(
            query=self._get_query(),
            dialect="standard",
            location=self._location,
            project_id=self._project_id,
            credentials=self._auth(),
        )
        ObjectStore.put(self._key, df)


class BigQueryFileDownload(BaseBigQuery):
    """
    Download query result as a csv file.

    This class saves BigQuery result as a temporary file in GCS, and then download.
    Download file could be a multiple if file size exceed 1GB.
    """

    def __init__(self):
        super().__init__()
        self._bucket = None
        self._dest_dir = None
        self._filename = None

    @property
    def bucket(self):
        return self._bucket

    @bucket.setter
    def bucket(self, bucket):
        self._bucket = bucket

    @property
    def dest_dir(self):
        return self._dest_dir

    @dest_dir.setter
    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, filename):
        self._filename = filename

    def execute(self, *args):
        super().execute()

        valid = EssentialParameters(
            self.__class__.__name__, [self._tblname, self._bucket, self._dest_dir]
        )
        valid()

        os.makedirs(self._dest_dir, exist_ok=True)

        gbq_client = self._bigquery_client()
        gbq_ref = gbq_client.dataset(self._dataset).table(self._tblname)

        gcs_client = self._gcs_client()
        gcs_bucket = gcs_client.get_bucket(self._bucket)

        ymd_hms = datetime.now().strftime("%Y%m%d%H%M%S%f")
        path = "%s-%s" % ("".join(random.choices(string.ascii_letters, k=8)), ymd_hms)
        prefix = "%s/%s/%s" % (self._dataset, self._tblname, path)

        # gsc dir -> gs://{bucket_name}/{dataset_name}/{table_name}/{XXXXXXXX}-{yyyyMMddHHmmssSSS}/*.csv.gz
        if self._filename:
            dest_gcs = "gs://%s/%s/%s*.csv.gz" % (self._bucket, prefix, self._filename)
        else:
            dest_gcs = "gs://%s/%s/*.csv.gz" % (self._bucket, prefix)

        # job config settings
        job_config = bigquery.ExtractJobConfig()
        job_config.compression = bigquery.Compression.GZIP
        job_config.desctination_format = bigquery.DestinationFormat.CSV

        # Execute query.
        job = gbq_client.extract_table(
            gbq_ref, dest_gcs, job_config=job_config, location=self._location
        )
        job.result()

        # Download from gcs
        for blob in gcs_bucket.list_blobs(prefix=prefix):
            dest = os.path.join(self._dest_dir, os.path.basename(blob.name))
            blob.download_to_filename(dest)

        # Cleanup temporary files
        for blob in gcs_bucket.list_blobs(prefix=prefix):
            blob.delete()


class GcsDownload(BaseGcs):
    """
    Download from GCS
    """

    def __init__(self):
        super().__init__()
        self._prefix = None
        self._delimiter = None
        self._src_pattern = None
        self._dest_dir = "."

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        self._prefix = prefix

    @property
    def delimiter(self):
        return self._delimiter

    @delimiter.setter
    def delimiter(self, delimiter):
        self._delimiter = delimiter

    @property
    def src_pattern(self):
        return self._src_pattern

    @src_pattern.setter
    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    @property
    def dest_dir(self):
        return self._dest_dir

    @dest_dir.setter
    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def execute(self, *args):
        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))
        super().execute()

        valid = EssentialParameters(self.__class__.__name__, [self._src_pattern])
        valid()

        client = self._gcs_client()
        bucket = client.get_bucket(self._bucket)
        dl_files = []
        for blob in bucket.list_blobs(prefix=self._prefix, delimiter=self._delimiter):
            r = re.compile(self._src_pattern)
            if not r.fullmatch(blob.name):
                continue
            dl_files.append(blob.name)
            blob.download_to_filename(
                os.path.join(self._dest_dir, os.path.basename(blob.name))
            )

        ObjectStore.put(self._step, dl_files)


class GcsDownloadFileDelete(BaseGcs):
    """
    Delete files downloaded from GCS by using cache
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))
        dl_files = ObjectStore.get(self._symbol)

        if len(dl_files) > 0:
            self._logger.info("Delete files %s" % dl_files)
            client = self._gcs_client()
            bucket = client.get_bucket(super().get_step_argument("bucket"))
            for blob in bucket.list_blobs(
                prefix=super().get_step_argument("prefix"),
                delimiter=super().get_step_argument("delimiter"),
            ):
                for dl_f in dl_files:
                    if dl_f == blob.name:
                        blob.delete()
                        break
        else:
            self._logger.info("No files to delete.")



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from abc import abstractmethod
from importlib import import_module
import os
import re
import subprocess

from cliboa.conf import env
from cliboa.core.validator import (
    ProjectDirectoryExistence,
    ScenarioFileExistence,
    DIScenarioFormat,
)
from cliboa.core.file_parser import YamlScenarioParser
from cliboa.core.scenario_queue import ScenarioQueue
from cliboa.scenario import *
from cliboa.util.lisboa_log import LisboaLog
from cliboa.util.cache import StepArgument
from cliboa.util.exception import *
from cliboa.util.http import FormAuth, BasicAuth

__all__ = ["YamlScenarioManager", "JsonScenarioManager"]


class ScenarioManager(object):
    """
    Management of scenario defined in files
    """

    # dependency injection keys in scenario.yml
    DI_KEYS = ["auth", "base_auth"]

    def __init__(self, cmd_args):
        self._logger = LisboaLog.get_logger(__name__)
        self._cmd_args = cmd_args
        self._pj_dir = os.path.join(env.PROJECT_DIR, cmd_args.project_name)
        self._pj_scenario_dir = os.path.join(
            env.PROJECT_DIR, cmd_args.project_name, env.SCENARIO_DIR_NAME
        )
        if cmd_args.format == "yaml":
            self._pj_scenario_file = (
                os.path.join(
                    env.PROJECT_DIR, cmd_args.project_name, env.SCENARIO_FILE_NAME
                )
                + ".yml"
            )
            self._cmn_scenario_file = (
                os.path.join(env.COMMON_DIR, env.SCENARIO_FILE_NAME) + ".yml"
            )
        else:
            self._pj_scenario_file = (
                os.path.join(
                    env.PROJECT_DIR, cmd_args.project_name, env.SCENARIO_FILE_NAME
                )
                + "."
                + cmd_args.format
            )
            self._cmn_scenario_file = (
                os.path.join(env.COMMON_DIR, env.SCENARIO_FILE_NAME)
                + "."
                + cmd_args.format
            )

        # key and var of dinamic variables
        self._dynamic_key_and_val = {}

    @abstractmethod
    def create_scenario_queue(self):
        """
        Create Scenario Queue
        """
        pass


class YamlScenarioManager(ScenarioManager):
    """
    Manage scenario with yml format
    """

    def create_scenario_queue(self):
        self._logger.info("Start to create scenario queue")

        # validation
        self.__valid_essential_dir()
        self.__valid_essential_files()

        # parse scenario.yml
        parser = YamlScenarioParser(self._pj_scenario_file, self._cmn_scenario_file)
        yaml_scenario_list = parser.parse()

        if yaml_scenario_list and isinstance(yaml_scenario_list, list):
            self.__invoke_steps(yaml_scenario_list)
        else:
            raise ScenarioFileInvalid("scenario.yml is invalid.")

        self._logger.info("Finish to create scenario queue")

    def __valid_essential_dir(self):
        """
        Project directory validation
        """
        valid_instance = ProjectDirectoryExistence()
        valid_instance(self._pj_dir)

    def __valid_essential_files(self):
        """
        Scenairo file validation
        """
        valid_instance = ScenarioFileExistence()
        valid_instance(self._pj_scenario_file)

    def __invoke_steps(self, yaml_scenario_list):
        """
        Create executable instance and push them to queue
        Args:
            yaml_scenario_list: parsed yaml list
        """
        self._logger.info("Start to invoke scenario")

        # Create queue to save step instances
        from cliboa.core.factory import StepQueueFactory

        q = StepQueueFactory.create("step")

        for s_dict in yaml_scenario_list:
            if "multi_process_count" in s_dict.keys():
                q.multi_proc_cnt = s_dict.get("multi_process_count")
                continue

            instances = []
            if "parallel" in s_dict.keys():
                for row in s_dict.get("parallel"):
                    instance = self.__create_instance(row, yaml_scenario_list)
                    setattr(instance, "logger", LisboaLog.get_logger(instance.__class__.__name__))
                    instances.append(instance)
                    StepArgument._put(row["step"], instance)
            else:
                instance = self.__create_instance(s_dict, yaml_scenario_list)
                setattr(instance, "logger", LisboaLog.get_logger(instance.__class__.__name__))
                instances.append(instance)
                StepArgument._put(s_dict["step"], instance)

            # Put instance to queue
            q.push(instances)

        # save queue to static area
        setattr(ScenarioQueue, "step_queue", q)
        self._logger.info("Finish to invoke scenario")

    def __create_instance(self, s_dict, yaml_scenario_list):
        cls_name = s_dict["class"]
        self._logger.debug("Create %s instance" % cls_name)

        if self.__is_custom_cls(cls_name) is True:
            from cliboa.core.factory import CustomInstanceFactory

            instance = CustomInstanceFactory.create(cls_name)
        else:
            cls = globals()[cls_name]
            instance = cls()

        base_args = ["step", "symbol", "parallel", "io"]
        for arg in base_args:
            setattr(instance, arg, s_dict.get(arg))

        cls_attrs_dict = {}
        if isinstance(yaml_scenario_list, list) and "arguments" in s_dict.keys():
            cls_attrs_dict = s_dict["arguments"]

        # loop and set class attribute
        di_key = None
        di_instance = None
        if cls_attrs_dict:
            cls_attrs_dict = self.__extract_with_vars(cls_attrs_dict)

            # check if the keys of dependency injection
            di_keys, di_instances = self.__create_di_instance(cls_attrs_dict)
            if di_keys and di_instances:
                di_keys_and_instances = zip(di_keys, di_instances)
                for di_key, di_instance in di_keys_and_instances:
                    self._logger.debug(
                        "Inject %s to %s object." % (di_instance, instance)
                    )
                    # Injection
                    setattr(instance, di_key, di_instance)
                    del cls_attrs_dict[di_key]

            for yaml_k, yaml_v in cls_attrs_dict.items():
                # if value includes {{ var }}, replace value specified by with_vars
                if isinstance(yaml_v, str):
                    pattern = re.compile(r"\{\{(.*)\}\}")
                    exists_var = pattern.search(yaml_v)
                    if exists_var:
                        var_name = exists_var.group(1).strip()
                        yaml_v = self.__replace_vars(yaml_v, var_name)

                # set attribute to step class
                setattr(instance, yaml_k, yaml_v)

        return instance

    def __is_custom_cls(self, cls_name):
        """
        The specified class in scenario.yml is a custom step class or not
        Args:
            cls_name: the specified step class in scenario.yml
        Return:
            True: custom step class
            False: default step class
        """
        custom_classes = env.COMMON_CUSTOM_CLASSES + env.PROJECT_CUSTOM_CLASSES
        for cc in custom_classes:
            split_cc = cc.split(".")
            custom_cls_name = split_cc[1]
            if cls_name == custom_cls_name:
                return True
        return False

    def __extract_with_vars(self, cls_attrs_dict):
        """
        If 'with_vars' exist in scenario.yml, extract and save to dictionary.
        After that, remove from list
        """
        # Extract with_vars if exists
        exists_with_vars = "with_vars" in cls_attrs_dict.keys()
        if exists_with_vars:
            variables = cls_attrs_dict["with_vars"]
            for yaml_k, yaml_v in variables.items():
                self._dynamic_key_and_val[yaml_k] = yaml_v
            del cls_attrs_dict["with_vars"]
        return cls_attrs_dict

    def __replace_vars(self, yaml_v, var_name):
        """
        Replace {{ var }} in string
        Args:
            yaml_v: replace target value
            var_name: means {{ var }} itself
        """
        cmd = self._dynamic_key_and_val[var_name]
        if not cmd:
            raise ScenarioFileInvalid(
                "scenario.yml is invalid. 'with_vars' definition against %s does not exist."
                % var_name
            )
        shell_output = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, shell=True
        ).communicate()[0]
        shell_output = shell_output.strip()
        # remove head byte string
        shell_output = re.sub("^b", "", str(shell_output))
        # remove '
        shell_output = re.sub("'", "", str(shell_output))
        return re.sub(r"\{\{(.*)\}\}", shell_output, yaml_v)

    def __create_di_instance(self, cls_attrs):
        """
        Create an instance to be injected
        Args:
            step class attributes
        Retrurn:
            DI attribute names, DI instances
        """
        di_keys = []
        di_instance = None
        di_instances = []
        di_params = None
        for k in cls_attrs.keys():
            if k in self.DI_KEYS:
                di_keys.append(k)
                di_params = cls_attrs.get(k)
                valid = DIScenarioFormat(k, di_params)
                valid()
                di_cls = di_params["class"]
                di_cls = globals()[di_cls]
                di_instance = di_cls()
                if di_instance:
                    self._logger.debug(
                        "An instance %s to be injected exists." % di_instance
                    )
                    del di_params["class"]

                # set attributes to instance
                if di_params:
                    for k, v in di_params.items():
                        setattr(di_instance, k, v)
                di_instances.append(di_instance)
        return di_keys, di_instances


class JsonScenarioManager(ScenarioManager):
    """
    Manage scenario with json format
    """

    def create_scenario_queue(self):
        """
        TODO: implement in the future
        """
        pass



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import ast
import csv
import os
import pandas
from datetime import datetime
from google.cloud import bigquery, storage
from google.oauth2 import service_account

from cliboa.core.validator import EssentialParameters
from cliboa.scenario.extract.file import FileRead
from cliboa.scenario.gcp import BaseBigQuery, BaseGcs
from cliboa.util.cache import ObjectStore
from cliboa.util.exception import InvalidFormat, InvalidFileCount, FileNotFound


class BigQueryCreate(BaseBigQuery):
    """
    Insert data into BigQuery table
    """

    # default bulk line count to change to dataframe object
    BULK_LINE_CNT = 10000

    # BigQuery insert mode
    REPLACE = "replace"
    APPEND = "append"

    def __init__(self):
        super().__init__()
        self.__table_schema = None
        self.__replace = True

    @property
    def table_schema(self):
        return self.__table_schema

    @table_schema.setter
    def table_schema(self, table_schema):
        self.__table_schema = table_schema

    @property
    def replace(self):
        return self.__replace

    @replace.setter
    def replace(self, replace):
        self.__replace = replace

    def execute(self, *args):
        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))

        super().execute()

        param_valid = EssentialParameters(
            self.__class__.__name__, [self.__table_schema]
        )
        param_valid()

        cache_list = []
        inserts = False
        # initial if_exists
        if_exists = self.REPLACE if self.__replace is True else self.APPEND
        with open(self._s.cache_file, "r", encoding="utf-8") as f:
            for i, l_str in enumerate(f):
                l_dict = ast.literal_eval(l_str)
                cache_list.append(l_dict)
                if len(cache_list) == self.BULK_LINE_CNT:
                    df = pandas.DataFrame(self.__create_insert_data(cache_list))
                    if inserts is True:
                        # if_exists after the first insert execution
                        if_exists = self.APPEND
                    dest_tbl = self._dataset + "." + self._tblname
                    self._logger.info(
                        "Start insert %s rows to %s" % (len(cache_list), dest_tbl)
                    )
                    df.to_gbq(
                        dest_tbl,
                        project_id=self._project_id,
                        if_exists=if_exists,
                        table_schema=self.__table_schema,
                        location=self._location,
                        credentials=self._auth(),
                    )
                    cache_list.clear()
                    inserts = True
            if len(cache_list) > 0:
                df = pandas.DataFrame(self.__create_insert_data(cache_list))
                if inserts is True:
                    # if_exists after the first insert execution
                    if_exists = self.APPEND
                dest_tbl = self._dataset + "." + self._tblname
                self._logger.info(
                    "Start insert %s rows to %s" % (len(cache_list), dest_tbl)
                )
                df.to_gbq(
                    dest_tbl,
                    project_id=self._project_id,
                    if_exists=if_exists,
                    table_schema=self.__table_schema,
                    location=self._location,
                    credentials=self._auth(),
                )
        self._s.remove()

    def __create_insert_data(self, cache_list):
        """
        Create insert data like the below.

        insert_data = {
            "column1": [1, 2],
            "column2": ["spam", "spam"],
            ...
        }

        Args
            cache_list: dictionary list of input cache
        """
        insert_data = {}
        columns = [name_and_type["name"] for name_and_type in self.__table_schema]
        for c in columns:
            v_list = [d.get(c) for d in cache_list]
            if not v_list:
                raise InvalidFormat(
                    "Specified column %s does not exist in an input file." % c
                )
            insert_data[c] = v_list
        return insert_data


class GcsFileUpload(BaseGcs):
    """
    Upload local files to GCS
    """

    def __init__(self):
        super().__init__()

        self._src_dir = None
        self._src_pattern = None
        self._dest_dir = ""

    @property
    def src_dir(self):
        return self._src_dir

    @src_dir.setter
    def src_dir(self, src_dir):
        self._src_dir = src_dir

    @property
    def src_pattern(self):
        return self._src_pattern

    @src_pattern.setter
    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    @property
    def dest_dir(self):
        return self._dest_dir

    @dest_dir.setter
    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    def execute(self, *args):
        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))

        super().execute()

        valid = EssentialParameters(
            self.__class__.__name__, [self._src_dir, self._src_pattern]
        )
        valid()

        gcs_client = self._gcs_client()
        bucket = gcs_client.get_bucket(self._bucket)
        files = super().get_target_files(self._src_dir, self._src_pattern)
        self._logger.info("Upload files %s" % files)
        for file in files:
            self._logger.info("Start upload %s" % file)
            blob = bucket.blob(os.path.join(self._dest_dir, os.path.basename(file)))
            blob.upload_from_filename(file)
            self._logger.info("Finish upload %s" % file)


class CsvReadBigQueryCreate(BaseBigQuery, FileRead):
    """
    Read csv and Insert data into BigQuery table
    """

    # default bulk line count to change to dataframe object
    BULK_LINE_CNT = 10000

    # BigQuery insert mode
    REPLACE = "replace"
    APPEND = "append"

    def __init__(self):
        super().__init__()
        self.__table_schema = None
        self.__replace = True
        self.__columns = []

    @property
    def table_schema(self):
        return self.__table_schema

    @table_schema.setter
    def table_schema(self, table_schema):
        self.__table_schema = table_schema

    @property
    def replace(self):
        return self.__replace

    @replace.setter
    def replace(self, replace):
        self.__replace = replace

    def execute(self, *args):
        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))

        BaseBigQuery.execute(self)
        FileRead.execute(self)

        param_valid = EssentialParameters(
            self.__class__.__name__, [self.__table_schema]
        )
        param_valid()

        files = super().get_target_files(self._src_dir, self._src_pattern)
        if len(files) > 1:
            raise InvalidFileCount("Input file must be only one.")
        if len(files) == 0:
            raise FileNotFound("The specified csv file not found.")

        insert_rows = []
        is_inserted = False
        # initial if_exists
        if_exists = self.REPLACE if self.__replace is True else self.APPEND
        self.__columns = [
            name_and_type["name"] for name_and_type in self.__table_schema
        ]
        with open(files[0], "r", encoding=self._encoding) as f:
            reader = csv.DictReader(f, delimiter=",")
            for r in reader:
                # extract only the specified columns
                row_dict = {}
                for c in self.__columns:
                    if not r.get(c):
                        continue
                    row_dict[c] = r.get(c)
                insert_rows.append(row_dict)

                if len(insert_rows) == self.BULK_LINE_CNT:
                    self.__exec_insert(insert_rows, is_inserted, if_exists)
                    insert_rows.clear()
                    is_inserted = True
            if len(insert_rows) > 0:
                self.__exec_insert(insert_rows, is_inserted, if_exists)

    def __exec_insert(self, insert_rows, is_inserted, if_exists):
        """
        Execute insert into a BigQuery table
        Args:
            insert_rows: rows to insert
            is_inserted: if the data is already inserted or not
            if_exists: replace or append
        """
        df = pandas.DataFrame(self.__format_insert_data(insert_rows))
        if is_inserted is True:
            # if_exists after the first insert execution
            if_exists = self.APPEND
        dest_tbl = self._dataset + "." + self._tblname
        self._logger.info("Start insert %s rows to %s" % (len(insert_rows), dest_tbl))
        df.to_gbq(
            dest_tbl,
            project_id=self._project_id,
            if_exists=if_exists,
            table_schema=self.__table_schema,
            location=self._location,
            credentials=self._auth(),
        )

    def __format_insert_data(self, insert_rows):
        """
        Format insert data to pass DataFrame as the below.

        insert_data = {
            "column1": [1, 2],
            "column2": ["spam", "spam"],
            ...
        }

        Args
            insert_rows: dictionary list of input cache
        """
        insert_data = {}
        for c in self.__columns:
            v_list = [d.get(c) for d in insert_rows]
            if not v_list:
                raise InvalidFormat(
                    "Specified column %s does not exist in an input file." % c
                )
            insert_data[c] = v_list
        return insert_data



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import cliboa
from cliboa.scenario.base import BaseStep
from cliboa.util.lisboa_log import LisboaLog


class SampleStep(BaseStep):
    """
    For unit test
    """

    def __init__(self):
        super().__init__()
        self._retry_count = 3
        self._logger = LisboaLog.get_logger(__name__)

    @property
    def retry_count(self):
        return self._retry_count

    @retry_count.setter
    def retry_count(self, retry_count):
        self._retry_count = retry_count

    def execute(self, *args):
        self._logger.info("Start %s" % self.__class__.__name__)
        self._logger.info("Finish %s" % self.__class__.__name__)


class SampleCustomStep(BaseStep):
    """
    For unit test
    """

    def __init__(self):
        BaseStep.__init__(self)
        self._retry_count = 3

    @property
    def retry_count(self):
        return self._retry_count

    @retry_count.setter
    def retry_count(self, retry_count):
        self._retry_count = retry_count

    def execute(self, *args):
        self._logger.info("Start %s" % self.__class__.__name__)
        self._logger.info("Finish %s" % self.__class__.__name__)



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from abc import abstractmethod
from google.oauth2 import service_account
import os

from cliboa.scenario.validator import (
    EssentialParameters,
    SqliteTableExistence,
    IOOutput,
)
from cliboa.util.cache import StorageIO, StepArgument
from cliboa.util.file import File
from cliboa.util.sqlite import SqliteAdapter
from cliboa.util.exception import *


class BaseStep(object):
    """
    Base class of all the step classes
    """

    def __init__(self):
        self._s = StorageIO()
        self._step = None
        self._symbol = None
        self._parallel = None
        self._io = None
        self._logger = None

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, step):
        self._step = step

    @property
    def symbol(self):
        return self._symbol

    @symbol.setter
    def symbol(self, symbol):
        self._symbol = symbol

    @property
    def parallel(self):
        return self._parallel

    @parallel.setter
    def parallel(self, parallel):
        self._parallel = parallel

    @property
    def io(self):
        return self._io

    @io.setter
    def io(self, io):
        self._io = io

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        self._logger = logger

    def trigger(self, *args):
        try:
            return self.execute(args)
        except Exception as e:
            return self._exception_dispatcher(e)

    @abstractmethod
    def execute(self, *args):
        pass

    def get_target_files(self, src_dir, src_pattern):
        """
        Search files either with regular expression
        """
        return File().get_target_files(src_dir, src_pattern)

    def get_step_argument(self, name):
        """
        Returns an argument from scenario.yaml definitions
        """
        sa = StepArgument.get(self._symbol)
        if sa:
            return sa.get(name)

    def _property_path_reader(self, src, encoding="utf-8"):
        """
        Returns an resource contents from the path if src starts with "path:", returns src if not
        """
        if src[:5].upper() == "PATH:":
            fpath = src[5:]
            if os.path.exists(fpath) is False:
                raise FileNotFound(src)
            with open(fpath, mode="r", encoding=encoding) as f:
                return f.read()
        return src

    def _exception_dispatcher(self, e):
        """
        Handle and dispath CliboaExceptions
        """
        # TODO Currently not doing anything
        raise e


class BaseSqlite(BaseStep):
    """
    Base class of all the sqlite classes
    """

    def __init__(self):
        super().__init__()
        self._sqlite_adptr = SqliteAdapter()
        self._dbname = None
        self._columns = []
        self._vacuum = False

    @property
    def dbname(self):
        return self._dbname

    @dbname.setter
    def dbname(self, dbname):
        self._dbname = dbname

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, columns):
        self._columns = columns

    @property
    def vacuum(self):
        return self._vacuum

    @vacuum.setter
    def vacuum(self, vacuum):
        self._vacuum = vacuum

    def execute(self, *args):
        # essential parameters check
        param_valid = EssentialParameters(self.__class__.__name__, [self._dbname])
        param_valid()

    def _dict_factory(self, cursor, row):
        d = {}
        for i, col in enumerate(cursor.description):
            d[col[0]] = row[i]
        return d

    def _close_database(self):
        """
        Disconnect sqlite database (execute vacuume if necessary)
        """
        self._sqlite_adptr.close()
        if self._vacuum is True:
            try:
                self._sqlite_adptr.connect(self._dbname)
                self._sqlite_adptr.execute("VACUUM")
            finally:
                self._sqlite_adptr.close()


class SqliteQueryExecute(BaseSqlite):
    """
    Execute only row query of insert, update or delete
    If would like to execute read query, use SqliteRead class
    """

    def __init__(self):
        super().__init__()
        self.__tblname = None
        self.__raw_query = None

    @property
    def tblname(self):
        return self.__tblname

    @tblname.setter
    def tblname(self, tblname):
        self.__tblname = tblname

    @property
    def raw_query(self):
        return self.__raw_query

    @raw_query.setter
    def raw_query(self, raw_query):
        self.__raw_query = raw_query

    def execute(self, *args):
        super().execute()
        self._sqlite_adptr.connect(self._dbname)
        self._sqlite_adptr.execute(self.__raw_query)
        self._sqlite_adptr.commit()


class Stdout(BaseStep):
    """
    Standard output for io: input
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        output_valid = IOOutput(self._io)
        output_valid()

        with open(self._s.cache_file, "r", encoding="utf-8") as f:
            for l in f:
                print(l)



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os

from cliboa.conf import env
from cliboa.util.lisboa_log import LisboaLog
from cliboa.util.exception import (
    CliboaException,
    SqliteInvalid,
    ScenarioFileInvalid,
)
from cliboa.util.sqlite import SqliteAdapter


class EssentialParameters(object):
    """
    Validation for the essential parameters of step class
    """

    def __init__(self, cls_name, param_list):
        """
        Args:
            cls_name: class name which has validation target parameters
            param_list: list of validation target parameters
        """
        self.__cls_name = cls_name
        self.__param_list = param_list

    def __call__(self):
        for p in self.__param_list:
            if not p:
                raise CliboaException(
                    "The essential parameter is not specified in %s." % self.__cls_name
                )


class SqliteTableExistence(object):
    """
    Validation for the table of sqlite
    """

    def __init__(self, dbname, tblname, returns_bool=False):
        """
        Args:
            dbname: database name
            tblname: table name
            returns_bool: return bool or not
        """
        self.__sqlite_adptr = SqliteAdapter()
        self.__dbname = dbname
        self.__tblname = tblname
        self.__returns_bool = returns_bool
        self._logger = LisboaLog.get_logger(__name__)

    def __call__(self):
        try:
            self.__sqlite_adptr.connect(self.__dbname)
            cur = self.__sqlite_adptr.fetch(
                'SELECT name FROM sqlite_master WHERE type="table" AND name="%s"'
                % self.__tblname
            )
            result = cur.fetchall()
            if self.__returns_bool is True:
                return True if result else False

            if not result and self.__returns_bool is False:
                raise SqliteInvalid("Sqlite table %s not found" % self.__tblname)

        finally:
            self.__sqlite_adptr.close()


class IOInput(object):
    """
    Validation for io: input
    """

    def __init__(self, io):
        self.__io = io

    def __call__(self):
        if self.__io != "input":
            raise ScenarioFileInvalid("io: input is not specified.")


class IOOutput(object):
    """
    Validation for io: input
    """

    def __init__(self, io):
        self.__io = io

    def __call__(self):
        if self.__io != "output":
            raise ScenarioFileInvalid("io: output is not specified.")



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.conf import env
from cliboa.core.manager import YamlScenarioManager, JsonScenarioManager
from cliboa.core.step_queue import *
from cliboa.core.strategy import *
from cliboa.scenario.sample_step import *


class ScenarioManagerFactory(object):
    """
    Create scenario manager instance
    """

    @staticmethod
    def create(cmd_args):
        """
        Create ScenarioManager Instance(YamlScenarioManager or JsonScenarioManager)
        Args:
            cmd_args: Command Line Arguments
        Returns:
            scenario manager instance
        """
        scenario_file_format = cmd_args.format
        scenario_manager_cls = scenario_file_format.capitalize() + "ScenarioManager"
        instance = globals()[scenario_manager_cls]
        return instance(cmd_args)


class StepQueueFactory(object):
    """
    Create scenario queue instance
    """

    @staticmethod
    def create(scenario_type):
        """
        Args:
            scenario_type: step or extract ot transform or load or None
        Returns:
            scenario queue instance
        """
        if scenario_type:
            queue_cls = scenario_type.capitalize() + "Queue"
        else:
            queue_cls = "StepQueue"

        instance = globals()[queue_cls]
        return instance()


class StepExecutorFactory(object):
    """
    Create step execution strategy instance
    """

    @staticmethod
    def create(obj):
        """
        Args:
            obj: queue which stores execution target steps
        Returns:
            step execution strategy instance
        """
        if len(obj) > 1:
            return MultiProcExecutor(obj)

        return SingleProcExecutor(obj)


class CustomInstanceFactory(object):
    """
    Import python module and create instance dynamically
    """

    @staticmethod
    def create(cls_name):
        custom_classes = env.COMMON_CUSTOM_CLASSES + env.PROJECT_CUSTOM_CLASSES
        custom_cls_list = [c for c in custom_classes if cls_name in c]
        custom_cls = " ".join(str(c) for c in custom_cls_list)
        components = custom_cls.split(".")
        mod = __import__(components[0])
        instance = getattr(mod, components[1])
        return instance()



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from cliboa.scenario.base import BaseSqlite
from cliboa.scenario.validator import (
    EssentialParameters,
    SqliteTableExistence,
    IOInput,
)


# deprecated
class SqliteRead(BaseSqlite):
    """
    Select * from the specifed table
    """

    def __init__(self):
        super().__init__()
        self._tblname = None
        self._raw_query = None

    @property
    def tblname(self):
        return self._tblname

    @tblname.setter
    def tblname(self, tblname):
        self._tblname = tblname

    @property
    def raw_query(self):
        return self._raw_query

    @raw_query.setter
    def raw_query(self, raw_query):
        self._raw_query = raw_query

    def execute(self, *args):
        super().execute()

        input_valid = IOInput(self._io)
        input_valid()

        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))

        param_valid = EssentialParameters(self.__class__.__name__, [self._tblname])
        param_valid()

        tbl_valid = SqliteTableExistence(self._dbname, self._tblname)
        tbl_valid()

        def dict_factory(cursor, row):
            d = {}
            for i, col in enumerate(cursor.description):
                d[col[0]] = row[i]
            return d

        self._sqlite_adptr.connect(self._dbname)
        cur = self._sqlite_adptr.fetch(sql=self.__get_query(), row_factory=dict_factory)
        for r in cur:
            self._s.save(r)

    def __get_query(self):
        """
        Get sql to read
        """
        if self._raw_query:
            return self._raw_query

        sql = ""
        if self._columns:
            select_columns = ",".join(map(str, self._columns))
            sql = "SELECT %s FROM %s" % (select_columns, self._tblname)
        else:
            sql = "SELECT * FROM %s" % self._tblname
        return sql


class SqliteReadRow(BaseSqlite):
    """
    Execute query.
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args):
        super().execute()
        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))

        self._sqlite_adptr.connect(self._dbname)
        try:
            cur = self._sqlite_adptr.fetch(sql=self._get_query(), row_factory=self._get_factory())
            self._callback_handler(cur)
        finally:
            self._sqlite_adptr.close()

    def _get_factory(self):
        """
        Default row factory (returns value as tuple) is used if factory is not set
        """
        return None

    def _get_query(self):
        raise NotImplementedError("Method 'get_query' must be implemented by subclass")

    def _callback_handler(self, cursor):
        raise NotImplementedError("Method 'callback_handler' must be implemented by subclass")



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import codecs
import csv
import os
import re


class File(object):
    def remove_csv_col(self, input_file, output_file, remains, enc="utf-8"):
        """
        Extract only the necessary columns from the CSV data and output a new CSV

        Args:
            input_file (str): Input csv file name
            output_file (str): Output csv file name
            remains (str[]): Columns which remain for new csv
            enc=utf-8 (str): encording
        """
        with codecs.open(input_file, mode="r", encoding=enc) as in_f, codecs.open(
            output_file, mode="w", encoding="utf-8"
        ) as out_f:
            reader = csv.DictReader(in_f)
            writer = csv.writer(out_f)
            writer.writerow(remains)

            for row in reader:
                contents = []
                for col in remains:
                    contents.append(row[col])
                writer.writerow(contents)
            out_f.flush()

    def get_target_files(self, src_dir, src_pattern, tree=True):
        """
        Get files which matches to the regular expression

        Args:
            src_dir (str): Directory to search
            src_pattern (str): Regular expression
            tree=True (bool): Set True(by default) to search files include sub directories.

        Returns:
            list: Matched file list
        """
        r = re.compile(src_pattern)
        target_files = []
        for dir, dirs, files in os.walk(src_dir):
            if tree is False and dir not in src_dir:
                continue
            for file in files:
                if r.fullmatch(file):
                    target_files.append(os.path.join(dir, file))
        return target_files



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from .base import Stdout, SqliteQueryExecute

from .extract.aws import S3Download
from .extract.file import CsvRead
from .extract.ftp import FtpDownload, FtpDownloadFileDelete
from .extract.sftp import (
    SftpDownload,
    SftpDelete,
    SftpDownloadFileDelete,
)
from .extract.http import HttpDownload
from .extract.gcp import (
    BigQueryReadCache,
    BigQueryFileDownload,
    GcsDownload,
    GcsDownloadFileDelete,
)
from .extract.sqlite import SqliteRead, SqliteReadRow

from .transform.file import (
    FileDecompress,
    FileCompress,
    CsvColsExtract,
    ColumnLengthAdjust,
    DateFormatConvert,
    ExcelConvert,
    CsvMerge,
    CsvHeaderConvert,
    FileDivide,
    FileRename,
)

from .load.aws import S3Upload
from .load.file import CsvWrite
from .load.sftp import SftpFileLoad, SftpUpload
from .load.sqlite import SqliteCreation, CsvReadSqliteCreate
from .load.gcp import BigQueryCreate, GcsFileUpload, CsvReadBigQueryCreate



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
import os


from cliboa.scenario.base import BaseStep
from cliboa.scenario.validator import EssentialParameters
from cliboa.util.http import Download


class HttpExtract(BaseStep):
    def __init__(self):
        super().__init__()
        self._src_url = None
        self._src_pattern = None
        self._dest_dir = None
        self._dest_pattern = None
        self._timeout = 30
        self._retry_count = 3

    @property
    def src_url(self):
        return self._src_url

    @src_url.setter
    def src_url(self, src_url):
        self._src_url = src_url

    @property
    def src_pattern(self):
        return self._src_pattern

    @src_pattern.setter
    def src_pattern(self, src_pattern):
        self._src_pattern = src_pattern

    @property
    def dest_dir(self):
        return self._dest_dir

    @dest_dir.setter
    def dest_dir(self, dest_dir):
        self._dest_dir = dest_dir

    @property
    def dest_pattern(self):
        return self._dest_pattern

    @dest_pattern.setter
    def dest_pattern(self, dest_pattern):
        self._dest_pattern = dest_pattern

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        self._timeout = timeout

    @property
    def retry_count(self):
        return self._retry_count

    @retry_count.setter
    def retry_count(self, retry_count):
        self._retry_count = retry_count

    def execute(self, *args):
        # essential parameters check
        valid = EssentialParameters(
            self.__class__.__name__, [self._src_url, self._src_pattern, self._dest_dir]
        )
        valid()


class HttpDownload(HttpExtract):
    def __init__(self):
        super().__init__()

    def execute(self, *args):
        for k, v in self.__dict__.items():
            self._logger.info("%s : %s" % (k, v))
        super().execute()

        os.makedirs(self._dest_dir, exist_ok=True)

        url = os.path.join(self._src_url, self._src_pattern)
        dest_path = (
            os.path.join(self._dest_dir, self._dest_pattern)
            if self._dest_pattern
            else os.path.join(self._dest_dir, self._src_pattern)
        )

        d = Download(url, dest_path, self._timeout, self._retry_count)
        d.execute()



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from abc import abstractmethod

from cliboa.util.lisboa_log import LisboaLog


class BaseListener(object):
    """
    Base listener for all the listener classes
    """

    def __init__(self):
        self._logger = LisboaLog.get_logger(__name__)


class ScenarioListener(BaseListener):
    """
    Listener for scenario
    """

    @abstractmethod
    def before_scenario(self, worker):
        """
        Update scenario execution status before scenario execution
        """
        pass

    @abstractmethod
    def after_scenario(self, worker):
        """
        Update scenario execution status after scenario completion
        """
        pass


class StepListener(BaseListener):
    """
    Listener for step
    """

    @abstractmethod
    def before_step(self, strategy):
        """
        Update step status before step execution
        """
        pass

    @abstractmethod
    def after_step(self, strategy):
        """
        Update step status after step
        """
        pass

    @abstractmethod
    def after_completion(self):
        """
        Update step status after step completion
        """
        pass


class ScenarioStatusListener(BaseListener):
    """
    Listener for scenario execution status
    """

    def before_scenario(self, worker):
        self._logger.info(
            "Start scenario execution. %s" % (worker.get_scenario_queue_status())
        )

    def after_scenario(self, worker):
        self._logger.info(
            "Finish scenario execution. %s" % (worker.get_scenario_queue_status())
        )


class StepStatusListener(BaseListener):
    """
    Listener for step execution status
    """

    def before_step(self, strategy):
        self._logger.info("Start step execution in %s." % strategy.get_queue_status())

    def after_step(self, strategy):
        self._logger.info("Finish step execution in %s." % strategy.get_queue_status())

    def after_completion(self):
        self._logger.info("Complete step execution.")



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from queue import Queue

__all__ = ["StepQueue"]


class StepQueue(Queue):
    """
    Queue for processing
    """
    DEFAULT_PARALLEL_CNT = 2

    def __init__(self):
        super().__init__()
        self.__multi_proc_cnt = self.DEFAULT_PARALLEL_CNT

    @property
    def multi_proc_cnt(self):
        return self.__multi_proc_cnt

    @multi_proc_cnt.setter
    def multi_proc_cnt(self, multi_proc_cnt):
        self.__multi_proc_cnt = multi_proc_cnt

    def push(self, instance):
        self.put(instance)

    def pop(self):
        return self.get()

    def peek(self):
        """
        Retrieves the next item, but does not remove.
        """
        return self.queue[0]

    def size(self):
        """
        Returns:
            queue size
        """
        return self.qsize()

    def is_empty(self):
        """
        Returns:
            if queue is empty: True
            if queue is not empty: False
        """
        return self.empty()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#

import argparse
from importlib import import_module
import os
import sys
from shutil import copyfile
import site


class CliboAdmin(object):
    """
    cliboa administrator command
    """

    def __init__(self, args):
        self.__args = args
        self.__cmn_dir = None
        self.__bin_dir = None

    def main(self):
        """
        Usage:
            cliboadmin init $directory_name
            cliboadmin create $project_name
        """
        if self.__args.option == "init":
            self.__init_pj(self.__args.dir_name)
            print(
                "Initialization of cliboa project '"
                + self.__args.dir_name
                + "' was successful."
            )
        elif self.__args.option == "create":
            self.__create_new_pj(self.__args.dir_name)
            print(
                "Adding a new project '" + self.__args.dir_name + "' was successfule."
            )

    def __init_pj(self, ini_dir):
        """
        Initialize program configuration
        """
        self.__create_ess_dirs(ini_dir)
        self.__create_ess_files(ini_dir)

    def __create_ess_dirs(self, ini_dir):
        """
        create essential directories
        """
        os.makedirs(ini_dir, exist_ok=False)
        self.__bin_dir = os.path.join(ini_dir, "bin")
        os.makedirs(self.__bin_dir, exist_ok=False)
        self.__cmn_dir = os.path.join(ini_dir, "common")
        os.makedirs(self.__cmn_dir, exist_ok=False)
        os.makedirs(os.path.join(self.__cmn_dir, "scenario"), exist_ok=False)
        os.makedirs(os.path.join(ini_dir, "conf"), exist_ok=False)
        os.makedirs(os.path.join(ini_dir, "logs"), exist_ok=False)
        os.makedirs(os.path.join(ini_dir, "project"), exist_ok=False)

    def __create_ess_files(self, ini_dir):
        """
        create essential files
        """
        lisboa_install_paths = site.getsitepackages()
        lisboa_install_path = (
            lisboa_install_paths[0]
            if os.path.exists(os.path.join(lisboa_install_paths[0], "cliboa"))
            else lisboa_install_paths[1]
        )

        run_cmd_path = os.path.join(
            lisboa_install_path, "cliboa", "template", "bin", "clibomanager.py"
        )
        copyfile(run_cmd_path, os.path.join(self.__bin_dir, "clibomanager.py"))

        requirements_path = os.path.join(
            lisboa_install_path, "cliboa/template", "requirements.txt"
        )
        copyfile(requirements_path, os.path.join(ini_dir, "requirements.txt"))

        cmn_env_path = os.path.join(
            lisboa_install_path, "cliboa", "conf", "default_environment.py"
        )
        copyfile(cmn_env_path, os.path.join(self.__cmn_dir, "environment.py"))

        cmn_scenario_path = os.path.join(ini_dir, "common", "scenario.yml")
        with open(cmn_scenario_path, "w") as yaml:
            yaml.write("scenario:" + "\n")

        cmn_ini_path = os.path.join(ini_dir, "common", "__init__.py")
        open(cmn_ini_path, "w").close()

    def __create_new_pj(self, new_pj_dir):
        """
        Create an individual project configuration
        """
        # check if being under lisboa project directory
        sys.path.append(os.getcwd())
        env = import_module("common.environment")

        # make essential directories and files
        os.makedirs(os.path.join("project", new_pj_dir), exist_ok=False)
        os.makedirs(os.path.join("project", new_pj_dir, "scenario"), exist_ok=False)
        with open(os.path.join("project", new_pj_dir, "scenario.yml"), "w") as yaml:
            yaml.write("scenario:" + "\n")


class CommandArgumentParser(object):
    def parse(self):
        """
        Parse cliboadmin arguments
        """
        parser = argparse.ArgumentParser()
        parser.add_argument("option", choices=["init", "create"], help="init or create")
        parser.add_argument(
            "dir_name", help="init $directory_name or create $project_directory_name"
        )
        return parser.parse_args()


def main():
    parser = CommandArgumentParser()
    args = parser.parse()
    admin = CliboAdmin(args)
    admin.main()


if __name__ == "__main__":
    main()



#
# Copyright 2019 BrainPad Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
from google.cloud import storage, bigquery
from google.oauth2 import service_account

from cliboa.scenario.base import BaseStep
from cliboa.scenario.validator import EssentialParameters


class BaseGcp(BaseStep):
    """
    Base class of Gcp usage.
    """

    def __init__(self):
        super().__init__()
        self._project_id = None
        self._credentials = None

    @property
    def project_id(self):
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        self._project_id = project_id

    @property
    def credentials(self):
        return self._credentials

    @credentials.setter
    def credentials(self, credentials):
        self._credentials = credentials

    def execute(self, *args):
        valid = EssentialParameters(self.__class__.__name__, [self._project_id])
        valid()

    def _auth(self):
        if self._credentials:
            return service_account.Credentials.from_service_account_file(self._credentials)

    def _bigquery_client(self):
        if self._credentials:
            return bigquery.Client.from_service_account_json(self._credentials)
        else:
            return bigquery.Client()

    def _gcs_client(self):
        if self._credentials:
            return storage.Client.from_service_account_json(self._credentials)
        else:
            return storage.Client()


class BaseBigQuery(BaseGcp):
    """
    Base class of BigQuery use.

    """

    def __init__(self):
        super().__init__()

        self._dataset = None
        self._tblname = None
        self._location = None

    @property
    def dataset(self):
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        self._dataset = dataset

    @property
    def tblname(self):
        return self._tblname

    @tblname.setter
    def tblname(self, tblname):
        self._tblname = tblname

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, location):
        self._location = location

    def execute(self, *args):
        super().execute()
        valid = EssentialParameters(
            self.__class__.__name__, [self._dataset, self._location]
        )


class BaseGcs(BaseGcp):
    """
    Base class of Gcs use.

    """

    def __init__(self):
        super().__init__()
        self._bucket = None

    @property
    def bucket(self):
        return self._bucket

    @bucket.setter
    def bucket(self, bucket):
        self._bucket = bucket

    def execute(self, *args):
        super().execute()
        valid = EssentialParameters(self.__class__.__name__, [self._bucket])
